# This file was auto-generated by salt's setup on Tuesday, 18 April 2017 @ 17:04:55 UTC.

from salt.version import SaltStackVersion

__saltstack_version__ = SaltStackVersion(2016, 11, 4, 0, '', 0, 0, None)
